import random
from typing import List
from ape.core.optimizer.optimizer_base import Optimizer
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem


class SampledFewshot(Optimizer):
    """
    A class for optimizing prompts by sampling few-shot examples from a dataset.

    This optimizer selects a subset of examples from the training set to use as few-shot examples
    in the prompt. The selection can be either random or sequential.

    Attributes:
        k (int): The maximum number of few-shot examples to include in the prompt. Defaults to 16.
    """

    k: int = 16

    async def optimize(
        self, student: Prompt, *, trainset: List[DatasetItem], randomize: bool = True, **kwargs
    ) -> Prompt:
        """
        Optimize the given prompt by sampling few-shot examples from the training set.

        Args:
            student (Prompt): The initial prompt to be optimized.
            trainset (List[DatasetItem]): The dataset to sample few-shot examples from.
            randomize (bool): Whether to randomly sample examples or take the first k. Defaults to True.

        Returns:
            Prompt: The optimized prompt with sampled few-shot examples.
        """
        self.student = student.reset_copy()
        self.trainset = trainset

        if len(self.trainset) == 0:
            return self.student

        if randomize:
            self.student.fewshot = random.sample(self.trainset, min(self.k, len(self.trainset)))
        else:
            self.student.fewshot = self.trainset[: min(self.k, len(self.trainset))]

        return self.student



import asyncio
import json
import random
from re import M
from typing import Any, Dict, List, Optional

from ape.common.evaluate.evaluate import Evaluate
from ape.common.prompt import Prompt
from ape.common.types import (
    ResponseFormat,
    DatasetItem,
    EvaluationResult,
)
from ape.common.utils import logger

from ape.core.core_prompts import ApeCorePrompts
from ape.core.proposer.dataset_summary_generator import create_dataset_summary
from ape.core.proposer.utils import extract_prompt
from ape.core.proposer.propose_base import Proposer

# TIPS for generating new instructions.
# TODO: Refine these tips.
TIPS = {
    "none": "Make it better",
    "one_side_first": "Try to optimize a certain part of the evaluation score first. For example, if the evaluation is based on both precision and recall, try to optimize for recall first. Be extreme at times.",
    "double_down": "Find what the prompt is already doing well and double down on it. For example, if the prompt already does a good job with precision, try to improve that even further with various strategies. Be extreme at times",
    "weakness_first": "Find what the prompt is already doing poorly based on the evaluation score and metric and try to improve that. For example, if the prompt has low recall, try to improve that with various strategies. Be extreme at times",
    "think": "Try to think about what the evaluation score and metric are actually measuring. Then, try to optimize for that. For example, if the evaluation score is calculated based on the average length of the generated instructions, try to optimize for that. Be extreme at times",
}


class EvaluationDrivenProposer(Proposer):
    def __init__(
        self,
        trainset: List[DatasetItem],
        use_dataset_summary=True,
        use_task_demos=True,
        use_tip=True,
        view_data_batch_size=10,
    ):
        """
        Initialize the GroundedProposer.

        Args:
            trainset (List[DatasetItem]): The training dataset.
            use_dataset_summary (bool, optional): Whether to use dataset summary. Defaults to True.
            use_task_demos (bool, optional): Whether to use task demonstrations. Defaults to True.
            use_tip (bool, optional): Whether to include tips. Defaults to True.
            view_data_batch_size (int, optional): The batch size for viewing data. Defaults to 10.
        """
        self.use_dataset_summary = use_dataset_summary
        self.use_task_demos = use_task_demos
        self.use_tip = use_tip

        self.trainset: List[DatasetItem] = trainset
        self.view_data_batch_size = view_data_batch_size
        self.describe_prompt = ApeCorePrompts.get("describe-prompt")
        self.generate_instructions = ApeCorePrompts.get("gen-instruction-with-eval")
        self.data_summary = None

    async def prepare_dataset_summary(
        self,
        view_data_batch_size=10,
    ):
        """
        Prepare a summary of the dataset.

        Args:
            view_data_batch_size (int, optional): The batch size for viewing data. Defaults to 10.
        """
        self.view_data_batch_size = view_data_batch_size
        logger.info("Preparing dataset summary")
        self.data_summary = await create_dataset_summary(
            trainset=self.trainset,
            view_data_batch_size=self.view_data_batch_size,
        )
        logger.info(f"DATA SUMMARY: {self.data_summary}")

    async def propose_prompts(
        self,
        N: int,
        T: float,
        evaluate: Evaluate,
        base_prompt: Optional[Prompt] = None,
        metric: Optional[str] = None,
        task_description: Optional[str] = None,
    ) -> List[Prompt]:
        """
        Propose a set of new instructions for the task based on specified criteria.

        Args:
            trial_logs (Dict[str, Any]): Logs from previous trials.
            N (int): Number of proposals to generate.
            T (float): Temperature for generation.
            evaluate (Evaluate): Evaluation function.
            base_prompt (Optional[Prompt], optional): Base prompt to start from. Defaults to None.
            metric (Optional[str], optional): Metric to use for evaluation. Defaults to None.

        Returns:
            List[Prompt]: A list of proposed prompts.
        """
        # This is not necessary. But I kept it here for now.
        if not self.data_summary and self.trainset:
            await self.prepare_dataset_summary()

        # This is the part where we evaluate the base prompt.
        base_evaluation_score, base_evaluation_result, global_metric_metadata = await evaluate(
            base_prompt, return_outputs=True, return_global_metric_metadata=True
        )

        evaluation_result = (
            "final score : "
            + str(base_evaluation_score)
            + "\nmetadata : "
            + str(global_metric_metadata)
        )
        _tasks = [
            self.generate_new_instruction(
                index=i,
                base_prompt=base_prompt,
                evaluation_result=evaluation_result,
                T=T,
                response_format=base_prompt.response_format,
                metric=metric,
                task_description=task_description,
            )
            for i in range(N)
        ]

        proposed_instructions = await asyncio.gather(*_tasks)

        return proposed_instructions

    async def generate_new_instruction(
        self,
        index: int,
        base_prompt: Prompt,
        # evaluation_score: str,
        evaluation_result: List[EvaluationResult],
        T: float,
        response_format: Optional[ResponseFormat] = None,
        metric: Optional[str] = None,
        task_description: Optional[str] = None,
        retry_count: int = 0,
    ) -> Prompt:
        """
        Generate a new instruction based on the base prompt and its evaluation result.

        Args:
            base_prompt (Prompt): The base prompt.
            evaluation_score (str): The score of the base prompt.
            evaluation_result (List[EvaluationResult]): The result of evaluating the base prompt.
            T (float): Temperature for generation.
            response_format (Optional[ResponseFormat], optional): Format for the response.
            metric (Optional[str], optional): Metric used for evaluation.
            retry_count (int): 현재 재시도 횟수

        Returns:
            Prompt: A new proposed prompt or the base prompt if generation fails after 3 attempts.
        """

        # Select a tip
        if self.use_tip:
            selected_tip = list(TIPS.values())[index % len(TIPS)]
        else:
            selected_tip = ""

        base_prompt_messages = [json.dumps(message) for message in base_prompt.messages]
        base_prompt_messages_str = "\n".join(base_prompt_messages)

        try:
            new_instruction_text = await self.generate_instructions(
                base_prompt=base_prompt_messages_str,
                evaluation_result=str(evaluation_result),
                evaluation_function=metric,
                tip=selected_tip,
                response_format=(
                    str(response_format.get("type"))
                    if isinstance(response_format, dict)
                    and response_format.get("type") != "json_schema"
                    else (
                        str(response_format.get("json_schema", {}).get("schema"))
                        if isinstance(response_format, dict)
                        else str(response_format)
                    )
                ),
                human_tip=task_description,
            )

            if "```prompt" not in new_instruction_text:
                new_instruction_text = "```prompt\n" + new_instruction_text
            extracted_prompt = extract_prompt(new_instruction_text)

            # Create a new Prompt object with the generated instruction
            new_prompt = Prompt.load(extracted_prompt)
            if not new_prompt.messages:
                raise ValueError("Generated prompt has no messages")
            new_prompt.model = base_prompt.model
            new_prompt.response_format = response_format
            new_prompt.name = "InstructNew"

        except Exception as e:
            logger.error(f"Error generating new instruction (attempt {retry_count + 1}): {e}")
            if retry_count < 2:  # 최대 3번 시도 (0, 1, 2)
                return await self.generate_new_instruction(
                    index=index,
                    base_prompt=base_prompt,
                    evaluation_result=evaluation_result,
                    T=T,
                    response_format=response_format,
                    metric=metric,
                    retry_count=retry_count + 1,
                )
            else:
                logger.error(
                    "Failed to generate new instruction after 3 attempts. Returning base prompt."
                )
                return base_prompt

        return new_prompt



import json
from typing import List
from ape.common.types import DatasetItem, ResponseFormat

from ape.core.core_prompts import ApeCorePrompts


async def process_batch(
    trainset: List[DatasetItem],
    batch_start: int,
    batch_end: int,
) -> str:
    """
    Process a batch of the dataset to generate observations.

    Args:
        trainset (List[DatasetItem]): The full training dataset.
        batch_start (int): The starting index of the batch.
        batch_end (int): The ending index of the batch.

    Returns:
        str: The generated observations for the batch.
    """
    descriptor = ApeCorePrompts.get("dataset-descriptor")

    # Format examples using a similar approach to format_fewshot
    formatted_examples = ""
    for idx, item in enumerate(trainset[batch_start:batch_end]):
        formatted_examples += f"### Demo {idx+1} ###\n"

        inputs, outputs = item.get("inputs", {}), item.get("outputs", {})

        formatted_examples += "**Inputs**\n"
        for key, value in inputs.items():
            formatted_examples += f"{key.capitalize()}:\n{value}\n"

        formatted_examples += f"**Outputs**\n{json.dumps(outputs, indent=2)}\n\n"

    output = await descriptor(
        lm_config=dict(temperature=1.0),
        examples=formatted_examples,
    )
    return output["observations"]


async def create_dataset_summary(
    trainset: List[DatasetItem],
    view_data_batch_size: int,
) -> str:
    """
    Create a summary of the dataset by processing a batch of it.

    Args:
        trainset (List[DatasetItem]): The full training dataset.
        view_data_batch_size (int): The maximum number of items to process.

    Returns:
        str: The generated observations summarizing the dataset.
    """
    upper_lim = min(len(trainset), view_data_batch_size)

    # Initial batch processing
    observations = await process_batch(trainset, 0, upper_lim)

    return observations



import os
from collections import defaultdict
import pickle
import random
from typing import Any, Dict, List, Optional, Tuple, Union
import numpy as np
import optuna
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from rich.prompt import Confirm
import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import partial

from ape.common.evaluate.evaluate import Evaluate
from ape.common.utils import logger
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem, ResponseFormat

from ..utils import run_async
from ape.core.optimizer.mipro.mipro_base import MIPROBase
from ape.core.optimizer.mipro.mipro_proposer import MIPROProposer
from ape.core.optimizer.utils import (
    create_n_fewshot_demo_sets,
    eval_candidate_prompt,
    get_prompt_with_highest_avg_score,
    reformat_prompt,
    save_candidate_prompt,
)

BOOTSTRAPPED_FEWSHOT_EXAMPLES_IN_CONTEXT: int = 3
LABELED_FEWSHOT_EXAMPLES_IN_CONTEXT: int = 0


class MIPRO(MIPROBase):
    """
    MIPRO (Multi-prompt Instruction PRoposal Optimizer) class for optimizing prompts.

    This class implements the MIPRO algorithm, which generates and evaluates multiple prompt
    candidates using optuna for hyperparameter optimization. It supports minibatch evaluation,
    few-shot examples, and various optimization strategies.

    For more details, see the MIPRO paper: https://arxiv.org/abs/2406.11695
    """

    def _get_batch_size(self, minibatch: bool, trainset: List[DatasetItem]) -> int:
        return self.minibatch_size if minibatch else len(trainset)

    def _display_warning_and_confirm(
        self,
        trainset: List[DatasetItem],
        max_steps: int,
        minibatch: bool,
        requires_permission_to_run: bool,
    ) -> bool:
        """
        Display a warning about projected LM calls and costs, and optionally ask for user confirmation.

        Args:
            trainset (List[DatasetItem]): The training dataset.
            max_steps (int): Maximum number of optimization steps.
            minibatch (bool): Whether minibatch optimization is used.
            requires_permission_to_run (bool): Whether user confirmation is required.

        Returns:
            bool: True if the user confirms or confirmation is not required, False otherwise.
        """
        console = Console()

        estimated_prompt_model_calls = 10 + self.num_candidates + 1

        if not minibatch:
            estimated_task_model_calls = len(trainset) * max_steps
            task_model_line = f"- Task Model: {len(trainset)} examples in train set * {max_steps} batches * # of LM calls in your program = ({estimated_task_model_calls} * # of LM calls in your program) task model calls"
        else:
            if self.update_prompt_after_full_eval:
                estimated_task_model_calls = self.minibatch_size * max_steps + (
                    len(trainset) * (max_steps // self.minibatch_full_eval_steps)
                )
                task_model_line = f"- Task Model: {self.minibatch_size} examples in minibatch * {max_steps} batches + {len(trainset)} examples in train set * {max_steps // self.minibatch_full_eval_steps} full evals = {estimated_task_model_calls} task model calls"
            else:
                estimated_task_model_calls = self.minibatch_size * max_steps
                task_model_line = f"- Task Model: {self.minibatch_size} examples in minibatch * {max_steps} batches = {estimated_task_model_calls} task model calls"

        warning_text = Text.from_markup(
            f"[bold yellow]WARNING: Projected Language Model (LM) Calls[/bold yellow]\n\n"
            f"Please be advised that based on the parameters you have set, the maximum number of LM calls is projected as follows:\n\n"
            f"[cyan]- Prompt Model: 10 data summarizer calls + {self.num_candidates} lm calls in program = {estimated_prompt_model_calls} prompt model calls[/cyan]\n"
            f"[cyan]{task_model_line}[/cyan]\n\n"
            f"[bold yellow]Estimated Cost Calculation:[/bold yellow]\n"
            f"Total Cost = (Number of calls to task model * (Avg Input Token Length per Call * Task Model Price per Input Token + Avg Output Token Length per Call * Task Model Price per Output Token) \n"
            f"            + (Number of calls to prompt model * (Avg Input Token Length per Call * Task Prompt Price per Input Token + Avg Output Token Length per Call * Prompt Model Price per Output Token).\n\n"
            f"For a preliminary estimate of potential costs, we recommend you perform your own calculations based on the task and prompt models you intend to use. "
            f"If the projected costs exceed your budget or expectations, you may consider:\n"
            f"- Reducing the number of trials (`max_steps`), the size of the trainset, or the number of LM calls in your program.\n"
            f"- Using a cheaper task model to optimize the prompt."
        )

        console.print(Panel(warning_text, title="Cost Warning", expand=False))

        if requires_permission_to_run:
            return Confirm.ask("Do you wish to continue?")
        return True

    async def optimize(
        self,
        student: Prompt,
        *,
        task_description: str = "",
        trainset: List[DatasetItem],
        testset: Optional[List[DatasetItem]] = None,
        max_steps: int = 30,
        max_bootstrapped_demos: int = 5,
        max_labeled_demos: int = 2,
        goal_score: float = 100,
        eval_kwargs: Optional[Dict[str, Any]] = None,
        seed: int = 9,
        minibatch: bool = True,
        requires_permission_to_run: bool = True,
        response_format: Optional[ResponseFormat] = None,
        log_dir: Optional[str] = None,
    ) -> Optional[Prompt]:
        """
        Optimize the given prompt using MIPRO (Multi-prompt Instruction PRoposal Optimizer).

        This method generates and evaluates multiple prompt candidates, using optuna for hyperparameter optimization.
        It supports minibatch evaluation, fewshot examples, and various optimization strategies.

        Returns:
            Optional[Prompt]: The best performing prompt, or None if optimization is aborted.
        """
        eval_kwargs = eval_kwargs or {}
        if not self._display_warning_and_confirm(
            trainset, max_steps, minibatch, requires_permission_to_run
        ):
            logger.info("Optimization aborted by the user.")
            return None

        random.seed(seed)
        np.random.seed(seed)
        testset = testset or trainset

        if response_format is not None:
            student = await reformat_prompt(prompt=student, response_format=response_format)

        response_format = student.response_format

        evaluate: Evaluate = self.evaluate or Evaluate(
            testset=testset, metric=self.metric, global_metric=self.global_metric, **eval_kwargs
        )

        max_bootstrapped_demos_for_candidate_gen: int = (
            BOOTSTRAPPED_FEWSHOT_EXAMPLES_IN_CONTEXT
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0
            else max_bootstrapped_demos
        )
        max_labeled_demos_for_candidate_gen: int = (
            LABELED_FEWSHOT_EXAMPLES_IN_CONTEXT
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0
            else max_labeled_demos
        )

        try:
            fewshot_candidates: Optional[List[List[DatasetItem]]] = (
                await create_n_fewshot_demo_sets(
                    student=student,
                    generate=self.generate,
                    num_candidate_sets=self.num_candidates,
                    trainset=trainset,
                    max_labeled_demos=max_labeled_demos_for_candidate_gen,
                    max_bootstrapped_demos=max_bootstrapped_demos_for_candidate_gen,
                    metric=self.metric,
                    teacher_settings=self.teacher_settings,
                    seed=seed,
                )
            )
            if log_dir:
                # if there is no directory, create it
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir)
                with open(os.path.join(log_dir, "fewshot_examples_to_save.pkl"), "wb") as file:
                    pickle.dump(fewshot_candidates, file)
        except Exception as e:
            logger.error(f"Error generating fewshot examples: {e}")
            logger.error("Running without fewshot examples.")
            fewshot_candidates = None
        logger.info("finished generating fewshot candidates")

        proposer: MIPROProposer = MIPROProposer(**self.model_dump())
        prompt_candidates: List[Prompt] = await proposer.generate_candidates(
            prompt=student,
            trainset=trainset,
            task_description=task_description,
            fewshot_candidates=fewshot_candidates,
            response_format=response_format,
        )
        logger.info("finished generating prompt candidates")

        if log_dir:
            with open(os.path.join(log_dir, "instructions_to_save.pkl"), "wb") as file:
                pickle.dump(prompt_candidates, file)

        instruction_candidates: List[List[Dict[str, str]]] = [
            prompt.messages for prompt in prompt_candidates
        ]

        if max_bootstrapped_demos == 0 and max_labeled_demos == 0:
            fewshot_candidates = None

        best_score: float = float("-inf")
        best_prompt: Optional[Prompt] = None
        trial_logs: Dict[int, Dict[str, Any]] = {}
        total_eval_calls: int = 0
        param_score_dict: Dict[str, List[Tuple[float, Prompt]]] = defaultdict(list)
        fully_evaled_param_combos: Dict[str, Dict[str, Union[Prompt, float]]] = {}

        def objective(trial: optuna.Trial) -> float:
            nonlocal best_prompt, best_score, trial_logs, total_eval_calls, param_score_dict, fully_evaled_param_combos

            logger.info(f"Starting trial num: {trial.number}")
            trial_logs[trial.number] = {}

            candidate_prompt: Prompt = student.deepcopy()

            instruction_idx: int = trial.suggest_categorical(
                "instruction", range(len(instruction_candidates))
            )
            chosen_params: List[int] = [instruction_idx]

            if fewshot_candidates:
                fewshot_idx: int = trial.suggest_categorical(
                    "fewshot", range(len(fewshot_candidates))
                )
                chosen_params.append(fewshot_idx)
                candidate_prompt.fewshot = fewshot_candidates[fewshot_idx]

            trial_logs[trial.number].update(
                {
                    "instruction": instruction_idx,
                    "fewshot": fewshot_idx if fewshot_candidates else None,
                }
            )

            candidate_prompt.messages = instruction_candidates[instruction_idx]

            trial_logs[trial.number]["prompt_path"] = save_candidate_prompt(
                candidate_prompt, log_dir, trial.number
            )

            batch_size: int = self._get_batch_size(minibatch, trainset)
            score: float = run_async(
                eval_candidate_prompt(batch_size, trainset, candidate_prompt, evaluate)
            )

            categorical_key: str = ",".join(map(str, chosen_params))
            param_score_dict[categorical_key].append((score, candidate_prompt))

            trial_logs[trial.number].update(
                {
                    "num_eval_calls": batch_size,
                    "full_eval": batch_size >= len(trainset),
                    "score": score,
                    "pruned": False,
                    "total_eval_calls_so_far": total_eval_calls + batch_size,
                }
            )
            total_eval_calls += batch_size

            if self.update_prompt_after_full_eval:
                if score > best_score and trial_logs[trial.number]["full_eval"] and not minibatch:
                    best_score = score
                    best_prompt = candidate_prompt.deepcopy()

                if minibatch and (
                    trial.number % self.minibatch_full_eval_steps == 0
                    or trial.number == max_steps - 1
                ):
                    trial_logs[trial.number]["mb_score"] = score
                    trial_logs[trial.number]["mb_prompt_path"] = trial_logs[trial.number][
                        "prompt_path"
                    ]

                    highest_mean_prompt, combo_key = get_prompt_with_highest_avg_score(
                        param_score_dict, fully_evaled_param_combos
                    )
                    full_train_score: float = run_async(
                        eval_candidate_prompt(
                            len(trainset), trainset, highest_mean_prompt, evaluate
                        )
                    )

                    fully_evaled_param_combos[combo_key] = {
                        "program": highest_mean_prompt,
                        "score": full_train_score,
                    }
                    total_eval_calls += len(trainset)
                    trial_logs[trial.number].update(
                        {
                            "total_eval_calls_so_far": total_eval_calls,
                            "full_eval": True,
                            "prompt_path": save_candidate_prompt(
                                prompt=highest_mean_prompt,
                                log_dir=log_dir,
                                trial_num=trial.number,
                                note="full_eval",
                            ),
                            "score": full_train_score,
                        }
                    )

                    if full_train_score > best_score:
                        best_score = full_train_score
                        best_prompt = highest_mean_prompt.deepcopy()
            else:
                if score > best_score:
                    best_score = score
                    best_prompt = candidate_prompt.deepcopy()

            if score >= goal_score:
                trial.study.stop()

            return score

        logger.info("starting optuna study")

        study: optuna.Study = optuna.create_study(
            direction="maximize",
            sampler=optuna.samplers.TPESampler(seed=seed, multivariate=True),
        )

        executor = ThreadPoolExecutor()

        try:
            # Use functools.partial to bind both the objective_args and n_trials
            bound_optimize = partial(
                study.optimize,
                func=objective,
                n_trials=max_steps,
            )

            # Run the bound optimize function in the executor
            await asyncio.get_event_loop().run_in_executor(executor, bound_optimize)
        except Exception as e:
            logger.error(f"Optuna study optimization failed: {e}")

        logger.info("finished optuna study")
        if best_prompt is not None and self.track_stats:
            best_prompt.metadata["trial_logs"] = trial_logs
            best_prompt.metadata["score"] = best_score
            best_prompt.metadata["total_eval_calls"] = total_eval_calls
        if log_dir:
            with open(os.path.join(log_dir, "best_prompt.prompt"), "w", encoding="utf-8") as f:
                f.write(best_prompt.dump())

            with open(os.path.join(log_dir, "optuna_study.pkl"), "wb") as file:
                pickle.dump(study, file)

        return best_prompt



import asyncio
import os
import random
from typing import Any, Dict, List, Optional, Set, Tuple

import numpy as np
from ape.common.evaluate.evaluate import Evaluate
from ape.common.generate.generate import Generate
from ape.common.generate.generate_base import BaseGenerate
from ape.common.metric.metric_base import BaseMetric
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem
from ape.common.types.response_format import ResponseFormat
from ape.common.utils import logger

from ape.core.optimizer.sampled_fewshot import SampledFewshot
from ape.core.optimizer.bootstrap_fewshot import BootstrapFewShot
from ape.core.proposer.utils import extract_prompt
from ape.core.core_prompts import ApeCorePrompts


_nest_asyncio_applied = False


async def reformat_prompt(prompt: Prompt, response_format: ResponseFormat) -> Prompt:
    """Reformat the prompt to be in XML style."""
    formatter_filename: str
    match response_format["type"]:
        case "json_object":
            formatter_filename = "reformat-prompt-json-object"
        case "json_schema":
            formatter_filename = "reformat-prompt-json-schema"
    formatter = ApeCorePrompts.get(formatter_filename)
    new_prompt: Prompt
    retry_count = 0
    logger.info(f"Reformatting prompt: {prompt.dump()}")
    while True:
        try:
            res = await formatter(prompt=prompt.dump())
            extracted = extract_prompt(res)
            if response_format["type"] == "json_object":
                if "json" not in extracted.lower():
                    raise ValueError("Reformatted prompt does not include the word 'JSON'")
            logger.info(f"Reformatted prompt: {extracted}")
            new_prompt = Prompt.load(extracted)
            new_prompt.name = prompt.name
            new_prompt.response_format = response_format
            break
        except Exception as e:
            logger.error(f"Error reformatting prompt: {e}. Retrying...")
            retry_count += 1
            if retry_count > 10:
                logger.error("Failed to reformat prompt after 3 retries")
                logger.error("Generated prompt:" + res)
                raise e

    # new_prompt.fewshot_config = prompt.fewshot_config # TODO: fix this more pretty way
    return new_prompt


def is_notebook():
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter
    except ImportError:
        return False  # IPython not installed


def run_async(coroutine):
    global _nest_asyncio_applied

    if not _nest_asyncio_applied:
        try:
            import nest_asyncio

            nest_asyncio.apply()
            _nest_asyncio_applied = True
        except ImportError:
            print("Please install nest_asyncio: !pip install nest_asyncio")
            raise

    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # If the loop is already running, create a new task and wait for it to complete
            task = asyncio.ensure_future(coroutine)
            return loop.run_until_complete(task)
        else:
            # If the loop is not running, use run_until_complete directly
            return loop.run_until_complete(coroutine)
    except RuntimeError:
        # If no event loop is present, use asyncio.run
        return asyncio.run(coroutine)


async def create_single_fewshot_demo_set(
    student: Prompt,
    trainset: List[DatasetItem],
    seed: int,
    max_labeled_demos: int,
    max_bootstrapped_demos: int,
    metric: BaseMetric,
    teacher_settings: dict,
    max_rounds: int,
    labeled_sample: bool,
    min_num_samples: int,
    metric_threshold: Any,
    teacher: Any,
    include_non_bootstrapped: bool,
    generate: BaseGenerate = Generate(),
) -> Prompt:
    trainset2 = list(trainset)

    if seed == -3 and include_non_bootstrapped:
        # zero-shot
        prompt2 = student.reset_copy()
    elif seed == -2 and max_labeled_demos > 0 and include_non_bootstrapped:
        # labels only
        optimizer = SampledFewshot(k=max_labeled_demos)
        prompt2 = await optimizer.optimize(student, trainset=trainset2, sample=labeled_sample)
    elif seed == -1:
        # unshuffled few-shot
        optimizer = BootstrapFewShot(
            generate=generate,
            metric=metric,
            max_bootstrapped_demos=max_bootstrapped_demos,
            max_labeled_demos=max_labeled_demos,
            teacher_settings=teacher_settings,
            max_rounds=max_rounds,
        )
        prompt2 = await optimizer.optimize(student, teacher=teacher, trainset=trainset2)
    else:
        # shuffled few-shot
        random.Random(seed).shuffle(trainset2)
        size = random.Random(seed).randint(min_num_samples, max_bootstrapped_demos)
        optimizer = BootstrapFewShot(
            generate=generate,
            metric=metric,
            metric_threshold=metric_threshold,
            max_bootstrapped_demos=size,
            max_labeled_demos=max_labeled_demos,
            teacher_settings=teacher_settings,
            max_rounds=max_rounds,
        )
        prompt2 = await optimizer.optimize(student, teacher=teacher, trainset=trainset2)

    return prompt2


async def create_n_fewshot_demo_sets(
    student: Prompt,
    num_candidate_sets: int,
    trainset: List[DatasetItem],
    max_labeled_demos: int,
    max_bootstrapped_demos: int,
    metric: BaseMetric,
    teacher_settings: dict,
    generate: BaseGenerate = Generate(),
    max_rounds=1,
    labeled_sample=True,
    min_num_samples=1,
    metric_threshold=None,
    teacher=None,
    include_non_bootstrapped=True,
    seed=0,
) -> List[List[DatasetItem]]:
    num_candidate_sets = max(num_candidate_sets, 3)
    random.Random(seed).shuffle(trainset)

    tasks = []
    for candidate_seed in range(-3, num_candidate_sets - 3):
        task = create_single_fewshot_demo_set(
            student=student,
            trainset=trainset,
            seed=candidate_seed,
            max_labeled_demos=max_labeled_demos,
            max_bootstrapped_demos=max_bootstrapped_demos,
            generate=generate,
            metric=metric,
            teacher_settings=teacher_settings,
            max_rounds=max_rounds,
            labeled_sample=labeled_sample,
            min_num_samples=min_num_samples,
            metric_threshold=metric_threshold,
            teacher=teacher,
            include_non_bootstrapped=include_non_bootstrapped,
        )
        tasks.append(task)

    max_workers = min(3, num_candidate_sets)
    semaphore = asyncio.Semaphore(max_workers)

    async def worker(task, i):
        async with semaphore:
            return await task

    fewshot_candidates = await asyncio.gather(*[worker(task, i) for i, task in enumerate(tasks)])

    return [
        prompt.fewshot for prompt in fewshot_candidates if prompt and hasattr(prompt, "fewshot")
    ]


def create_minibatch(trainset, batch_size=50):
    """Create a minibatch from the trainset."""

    # Ensure batch_size isn't larger than the size of the dataset
    batch_size = min(batch_size, len(trainset))

    # Randomly sample indices for the mini-batch
    sampled_indices = random.sample(range(len(trainset)), batch_size)

    # Create the mini-batch using the sampled indices
    minibatch = [trainset[i] for i in sampled_indices]

    return minibatch


async def eval_candidate_prompt(
    batch_size: int,
    trainset: List[DatasetItem],
    candidate_prompt: Prompt,
    evaluate: Evaluate,
):
    """Evaluate a candidate program on the trainset, using the specified batch size."""
    try:
        # Evaluate on the full trainset
        if batch_size >= len(trainset):
            score = await evaluate(candidate_prompt, testset=trainset, display_table=0)
        # Or evaluate on a minibatch
        else:
            score = await evaluate(
                candidate_prompt,
                testset=create_minibatch(trainset, batch_size),
                display_table=0,
            )
        if isinstance(score, tuple):
            score = score[0]

        return score
    except Exception as exc:
        logger.error(f"Error evaluating candidate prompt: {exc}")
        return 0


def save_candidate_prompt(prompt: Prompt, log_dir: Optional[str], trial_num: int, note=None):
    """Save the candidate prompt to the log directory."""

    if log_dir is None:
        return None

    # Ensure the directory exists
    eval_programs_dir = os.path.join(log_dir, "evaluated_prompts")
    os.makedirs(eval_programs_dir, exist_ok=True)

    # Define the save path for the program
    if note:
        save_path = os.path.join(eval_programs_dir, f"prompt_{trial_num}_{note}.prompt")
    else:
        save_path = os.path.join(eval_programs_dir, f"prompt_{trial_num}.prompt")

    # Save the prompt
    with open(save_path, "w") as f:
        f.write(prompt.dump())
    return save_path


def get_prompt_with_highest_avg_score(
    param_score_dict: Dict[str, List[Tuple[float, Prompt]]],
    fully_evaled_param_combos: Set[str],
) -> Tuple[Prompt, str]:
    """
    Returns the prompt with the highest average score from the batches evaluated so far.

    This function is used as a helper for Bayesian and minibatching optimizers.

    Args:
        param_score_dict: A dictionary mapping parameter combinations to lists of (score, prompt) tuples.
        fully_evaled_param_combos: A set of parameter combinations that have been fully evaluated.

    Returns:
        A tuple containing the best prompt and its corresponding parameter combination key.
    """
    results = [
        (key, np.mean([v[0] for v in values]), values[0][1])
        for key, values in param_score_dict.items()
    ]

    sorted_results = sorted(results, key=lambda x: x[1], reverse=True)

    for key, mean, prompt in sorted_results:
        if key not in fully_evaled_param_combos:
            logger.info(f"Best Combination: {key} with Mean = {mean:.4f}")
            return prompt, key

    # If all combinations are fully evaluated, return the overall best
    best_prompt, best_key = sorted_results[0][2], sorted_results[0][0]
    logger.warning("All parameter combinations fully evaluated. Returning overall best.")
    return best_prompt, best_key


async def find_best_fewshot(
    student: Prompt,
    num_candidate_sets: int,
    trainset: List[DatasetItem],
    max_labeled_demos: int,
    max_bootstrapped_demos: int,
    metric: BaseMetric,
    teacher_settings: dict,
    evaluate: Evaluate,
    max_rounds=1,
    labeled_sample=True,
    min_num_samples=1,
    metric_threshold=None,
    teacher=None,
    include_non_bootstrapped=True,
    batch_size=25,
    seed=0,
) -> Tuple[List[DatasetItem], float]:
    # Generate candidate few-shot sets
    fewshot_candidates = await create_n_fewshot_demo_sets(
        student=student,
        num_candidate_sets=num_candidate_sets,
        trainset=trainset,
        max_labeled_demos=max_labeled_demos,
        max_bootstrapped_demos=max_bootstrapped_demos,
        metric=metric,
        teacher_settings=teacher_settings,
        max_rounds=max_rounds,
        labeled_sample=labeled_sample,
        min_num_samples=min_num_samples,
        metric_threshold=metric_threshold,
        teacher=teacher,
        include_non_bootstrapped=include_non_bootstrapped,
        seed=seed,
    )

    async def evaluate_candidate(fewshot, semaphore):
        async with semaphore:
            candidate_prompt = student.deepcopy()
            candidate_prompt.fewshot = fewshot
            score = await eval_candidate_prompt(
                batch_size=min(batch_size, len(trainset)),  # Evaluate on minibatch
                trainset=trainset,
                candidate_prompt=candidate_prompt,
                evaluate=evaluate,
            )
            return (score, candidate_prompt, fewshot)

    max_concurrent = 5
    semaphore = asyncio.Semaphore(max_concurrent)

    results = await asyncio.gather(
        *[evaluate_candidate(fewshot, semaphore) for fewshot in fewshot_candidates]
    )

    best_score, best_prompt, best_fewshot = max(results, key=lambda x: x[0])

    return best_fewshot, best_score


# Example usage:
# best_prompt, best_score = await find_best_fewshot(
#     student=student_prompt,
#     num_candidate_sets=5,
#     trainset=trainset,
#     max_labeled_demos=3,
#     max_bootstrapped_demos=5,
#     metric=some_metric,
#     teacher_settings=teacher_settings,
#     evaluate=evaluate_function,
#     teacher=teacher_model
# )
# print(f"Best few-shot prompt score: {best_score}")
# print(f"Best few-shot prompt: {best_prompt.dump()}")



from abc import ABC, abstractmethod


class Proposer(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def propose_prompts(self):
        pass

    async def propose_one(self):
        pass



import os
from ape.common.prompt import Prompt
import promptfile as pf

current_dir = os.path.dirname(os.path.abspath(__file__))


if os.path.exists(current_dir):
    ApeCorePrompts = pf.Client(base_path=current_dir, prompt_class=Prompt)
else:
    raise FileNotFoundError(f"Prompts directory not found at {current_dir}")



from collections import defaultdict
import os
import pickle
import random
import inspect
from typing import Any, Dict, List, Optional, Tuple, Union
import numpy as np
import optuna
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from rich.prompt import Confirm

from ape.common.utils import logger
from ape.common.evaluate.evaluate import Evaluate
from ape.common.metric import (
    BaseMetric,
    CosineSimilarityMetric,
    JsonMatchMetric,
    SemanticF1Metric,
)
from ape.common.global_metric import BaseGlobalMetric
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem, ResponseFormat

from ape.core.core_prompts import ApeCorePrompts
from .mipro_base import MIPROBase
from .mipro_proposer import MIPROProposer
from ..utils import (
    run_async,
    eval_candidate_prompt,
    find_best_fewshot,
    reformat_prompt,
    save_candidate_prompt,
)
from ...proposer.evaluation_driven_proposer import EvaluationDrivenProposer
from ...proposer.utils import extract_prompt

BOOTSTRAPPED_FEWSHOT_EXAMPLES_IN_CONTEXT: int = 3
LABELED_FEWSHOT_EXAMPLES_IN_CONTEXT: int = 0


# Instruction Generation. Later Add Optuna and GroundedProposer to do (N+1) x (M+1) search.
class MIPROInstruct(MIPROBase):
    """
    MIPROInstruct class for optimizing prompts.

    This class is a modified vesion of MIPRO with a focus on generating better instructions. It supports minibatch evaluation,
    few-shot examples, and various optimization strategies.
    """

    gen_metric_description: Prompt = ApeCorePrompts.get("gen-metric-description")
    gen_metric_description_with_global_metric: Prompt = ApeCorePrompts.get(
        "gen-metric-description-with-global-metric"
    )
    merge_prompts: Prompt = ApeCorePrompts.get("gen-merged-prompt")

    def __init__(
        self,
        metric: BaseMetric,
        global_metric: Optional[BaseGlobalMetric] = None,
        teacher_settings: Dict[str, Any] = {},
        num_candidates: int = 10,
        init_temperature: float = 1.0,
        verbose: bool = False,
        track_stats: bool = True,
        log_dir: Optional[str] = None,
        view_data_batch_size: int = 10,
        minibatch_size: int = 25,
        minibatch_full_eval_steps: int = 10,
    ):
        """
        Initialize the MIPROInstruct optimizer.

        Args:
            metric (BaseMetric): The metric object used for evaluation.
            global_metric (Optional[BaseGlobalMetric]): The global metric object for overall evaluation.
            teacher_settings (Dict[str, Any]): Settings for the teacher model.
            num_candidates (int): Number of candidate prompts to generate.
            init_temperature (float): Initial temperature for sampling.
            verbose (bool): Whether to print verbose output.
            track_stats (bool): Whether to track statistics during optimization.
            log_dir (Optional[str]): Directory for logging.
            view_data_batch_size (int): Batch size for viewing data.
            minibatch_size (int): Size of minibatches for evaluation.
            update_prompt_after_full_eval (bool): Whether to update the best prompt after full evaluation.
            minibatch_full_eval_steps (int): Number of steps for full evaluation on minibatches.
        """

        super().__init__(
            teacher_settings=teacher_settings,
            num_candidates=num_candidates,
            metric=metric,
            global_metric=global_metric,
            init_temperature=init_temperature,
            verbose=verbose,
            track_stats=track_stats,
            log_dir=log_dir,
            view_data_batch_size=view_data_batch_size,
            minibatch_size=minibatch_size,
            minibatch_full_eval_steps=minibatch_full_eval_steps,
        )

    def _get_batch_size(self, minibatch: bool, trainset: List[DatasetItem]) -> int:
        return self.minibatch_size if minibatch else len(trainset)

    def _display_warning_and_confirm(
        self,
        trainset: List[DatasetItem],
        testset: List[DatasetItem],
        max_steps: int,
        minibatch: bool,
        requires_permission_to_run: bool,
    ) -> bool:
        """
        Display a warning about projected LM calls and costs, and optionally ask for user confirmation.

        Args:
            trainset (List[DatasetItem]): The training dataset.
            max_steps (int): Maximum number of optimization steps.
            minibatch (bool): Whether minibatch optimization is used.
            requires_permission_to_run (bool): Whether user confirmation is required.

        Returns:
            bool: True if the user confirms or confirmation is not required, False otherwise.
        """
        console = Console()

        estimated_prompt_model_calls = 10 + self.num_candidates + 1

        # TODO: Modify to match the actual number of LM calls in the program.
        if not minibatch:
            estimated_task_model_calls = len(testset) * max_steps
            task_model_line = f"- Task Model: {len(testset)} examples in test set * {max_steps} batches * # of LM calls in your program = ({estimated_task_model_calls} * # of LM calls in your program) task model calls"
        else:
            # if self.update_prompt_after_full_eval:
            #     estimated_task_model_calls = self.minibatch_size * max_steps + (
            #         len(trainset) * (max_steps // self.minibatch_full_eval_steps)
            #     )
            #     task_model_line = f"- Task Model: {self.minibatch_size} examples in minibatch * {max_steps} batches + {len(trainset)} examples in train set * {max_steps // self.minibatch_full_eval_steps} full evals = {estimated_task_model_calls} task model calls"
            # else:
            estimated_task_model_calls = self.minibatch_size * max_steps
            task_model_line = f"- Task Model: {self.minibatch_size} examples in minibatch * {max_steps} batches = {estimated_task_model_calls} task model calls"

        warning_text = Text.from_markup(
            f"[bold yellow]WARNING: Projected Language Model (LM) Calls[/bold yellow]\n\n"
            f"Please be advised that based on the parameters you have set, the maximum number of LM calls is projected as follows:\n\n"
            f"[cyan]- Prompt Model: 10 data summarizer calls + {self.num_candidates} lm calls in program = {estimated_prompt_model_calls} prompt model calls[/cyan]\n"
            f"[cyan]{task_model_line}[/cyan]\n\n"
            f"[bold yellow]Estimated Cost Calculation:[/bold yellow]\n"
            f"Total Cost = (Number of calls to task model * (Avg Input Token Length per Call * Task Model Price per Input Token + Avg Output Token Length per Call * Task Model Price per Output Token) \n"
            f"            + (Number of calls to prompt model * (Avg Input Token Length per Call * Task Prompt Price per Input Token + Avg Output Token Length per Call * Prompt Model Price per Output Token).\n\n"
            f"For a preliminary estimate of potential costs, we recommend you perform your own calculations based on the task and prompt models you intend to use. "
            f"If the projected costs exceed your budget or expectations, you may consider:\n"
            f"- Reducing the number of trials (`max_steps`), the size of the trainset, or the number of LM calls in your program.\n"
            f"- Using a cheaper task model to optimize the prompt."
        )

        console.print(Panel(warning_text, title="Cost Warning", expand=False))

        if requires_permission_to_run:
            return Confirm.ask("Do you wish to continue?")
        return True

    async def generate_metric_description(self) -> str:
        """
        Generate a description of the metric.
        """
        try:
            if isinstance(self.metric, CosineSimilarityMetric):
                metric_str = "Measures how similar the predicted text is to the correct answer by comparing their vector representations. A higher score means the prediction is more similar to the gold."
            elif isinstance(self.metric, JsonMatchMetric):
                metric_str = "Compares JSON format predictions with the correct JSON answer. It checks if each key-value pair in the prediction matches the ground truth exactly. The score reflects how many pairs match correctly."
            elif isinstance(self.metric, SemanticF1Metric):
                metric_str = """\
        Evaluates how well the prediction captures the meaning of the correct answer:
        1. Extracts key statements from both the prediction and ground truth.
        2. Checks how many statements from the prediction are found in the ground truth (Precision).
        3. Checks how many statements from the ground truth are found in the prediction (Recall).
        4. Calculates the F1 score, which balances Precision and Recall. A higher score indicates better semantic matching."""
            else:
                compute_function = getattr(self.metric, "compute", None)
                compute_function_source_code = inspect.getsource(compute_function)

                if self.global_metric:
                    global_metric_compute_function = getattr(self.global_metric, "compute", None)
                    global_metric_compute_function_source_code = inspect.getsource(
                        global_metric_compute_function
                    )

                    # get Prompt gen-metric-description-with-global-metric.prompt
                    metric_str = await self.gen_metric_description_with_global_metric(
                        **{
                            "metric_sourcecode": compute_function_source_code,
                            "global_metric_sourcecode": global_metric_compute_function_source_code,
                        }
                    )

                else:
                    metric_str = await self.gen_metric_description(
                        **{
                            "metric_sourcecode": compute_function_source_code,
                        }
                    )
                return metric_str
        except Exception as e:
            logger.error(f"Error generating metric description: {e}")
            return ""

    async def generate_instructions(
        self,
        student: Prompt,
        *,
        trainset: List[DatasetItem],
        testset: Optional[List[DatasetItem]] = None,
        eval_kwargs: Optional[Dict[str, Any]] = {},
        seed: int = 9,
        response_format: Optional[ResponseFormat] = None,
        log_dir: str,
    ) -> Optional[Prompt]:
        """
        Optimize the given prompt using MIPRO (Multi-prompt Instruction PRoposal Optimizer).

        This method generates and evaluates multiple prompt candidates, using optuna for hyperparameter optimization.
        It supports minibatch evaluation, fewshot examples, and various optimization strategies.

        Returns:
            Optional[Prompt]: The best performing prompt, or None if optimization is aborted.
        """

        random.seed(seed)
        np.random.seed(seed)
        testset = testset or trainset

        if response_format is not None:
            student = await reformat_prompt(prompt=student, response_format=response_format)

        metric_str = await self.generate_metric_description()

        evaluate: Evaluate = Evaluate(
            testset=testset,
            metric=self.metric,
            global_metric=self.global_metric,
            **eval_kwargs,
        )

        # This is the proposer.
        proposer = EvaluationDrivenProposer(
            trainset=trainset,
            view_data_batch_size=self.view_data_batch_size,
        )

        logger.info(f"Generating {self.num_candidates} explore instruction candidates")

        # Generate N candidates using EvaluationDrivenProposer.
        explore_prompt_candidates = await proposer.propose_prompts(
            base_prompt=student,
            N=self.num_candidates,
            T=self.init_temperature,
            evaluate=evaluate,
            metric=metric_str,
        )

        if log_dir:
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            with open(os.path.join(log_dir, "instructions_to_save.pkl"), "wb") as file:
                pickle.dump(explore_prompt_candidates, file)

        instruction_candidates: List[List[Dict[str, str]]] = [
            prompt.messages for prompt in explore_prompt_candidates
        ]

        return instruction_candidates

    async def optimize(
        self,
        student: Prompt,
        *,
        task_description: str = "",
        trainset: List[DatasetItem],
        testset: Optional[List[DatasetItem]] = None,
        max_steps: int = 30,
        max_bootstrapped_demos: int = 5,
        max_labeled_demos: int = 2,
        goal_score: float = 100,
        eval_kwargs: Optional[Dict[str, Any]] = None,
        seed: int = 9,
        minibatch: bool = True,
        requires_permission_to_run: bool = True,
        response_format: Optional[ResponseFormat] = None,
        log_dir: str,
        metric_threshold: Optional[float] = None,
    ) -> Optional[Prompt]:
        """
        Optimize the given prompt using MIPRO (Multi-prompt Instruction PRoposal Optimizer).

        This method generates and evaluates multiple prompt candidates, using optuna for hyperparameter optimization.
        It supports minibatch evaluation, fewshot examples, and various optimization strategies.

        Returns:
            Optional[Prompt]: The best performing prompt, or None if optimization is aborted.
        """
        eval_kwargs = eval_kwargs or {}
        if not self._display_warning_and_confirm(
            trainset, testset, max_steps, minibatch, requires_permission_to_run
        ):
            logger.info("Optimization aborted by the user.")
            return None

        random.seed(seed)
        np.random.seed(seed)
        if testset is None:
            testset = trainset

        if response_format is not None:
            student = await reformat_prompt(prompt=student, response_format=response_format)

        max_bootstrapped_demos_for_candidate_gen: int = (
            BOOTSTRAPPED_FEWSHOT_EXAMPLES_IN_CONTEXT
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0
            else max_bootstrapped_demos
        )
        max_labeled_demos_for_candidate_gen: int = (
            LABELED_FEWSHOT_EXAMPLES_IN_CONTEXT
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0
            else max_labeled_demos
        )

        trainset_evaluate: Evaluate = Evaluate(
            testset=trainset, metric=self.metric, global_metric=self.global_metric, **eval_kwargs
        )

        testset_evaluate: Evaluate = Evaluate(
            testset=testset, metric=self.metric, global_metric=self.global_metric, **eval_kwargs
        )

        logger.info("Start Find Best Fewshot")

        # find best few-shot
        best_fewshot, best_score = await find_best_fewshot(
            student=student,
            num_candidate_sets=self.num_candidates,
            trainset=trainset,
            max_labeled_demos=max_labeled_demos_for_candidate_gen,
            max_bootstrapped_demos=max_bootstrapped_demos_for_candidate_gen,
            metric=self.metric,
            teacher_settings=self.teacher_settings,
            seed=seed,
            evaluate=testset_evaluate,
            batch_size=self.minibatch_size,
            metric_threshold=metric_threshold,
        )

        logger.info("Start Propose Instruction Candidates from Evaluation Result")

        score_based_proposer = EvaluationDrivenProposer(
            trainset=trainset,
            view_data_batch_size=self.view_data_batch_size,
            use_tip=True,
        )

        logger.info(f"Generating {self.num_candidates} explore instruction candidates")

        metric_str = await self.generate_metric_description()

        # Generate N candidates using EvaluationDrivenProposer.
        score_based_prompt_candidates = await score_based_proposer.propose_prompts(
            base_prompt=student,
            N=self.num_candidates,
            T=self.init_temperature,
            evaluate=trainset_evaluate,
            metric=metric_str,
            task_description=task_description,
        )

        score_based_instruction_candidates: List[List[Dict[str, str]]] = [
            prompt.messages for prompt in score_based_prompt_candidates
        ]

        if log_dir:
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            with open(os.path.join(log_dir, "score_based_instructions_to_save.pkl"), "wb") as file:
                pickle.dump(score_based_prompt_candidates, file)

        logger.info("Start Propose Instruction Candidates from Prompt Engineering Techniques")

        format_based_proposer: MIPROProposer = MIPROProposer(**self.model_dump())
        format_based_prompt_candidates: List[Prompt] = (
            await format_based_proposer.generate_candidates(
                prompt=student,
                trainset=trainset,
                task_description=task_description,
                response_format=response_format,
            )
        )

        if log_dir:
            with open(os.path.join(log_dir, "format_based_instructions_to_save.pkl"), "wb") as file:
                pickle.dump(format_based_prompt_candidates, file)

        format_based_instruction_candidates: List[List[Dict[str, str]]] = [
            prompt.messages for prompt in format_based_prompt_candidates
        ]

        best_score: float = float("-inf")
        best_prompt: Optional[Prompt] = None
        trial_logs: Dict[int, Dict[str, Any]] = {}
        total_eval_calls: int = 0
        param_score_dict: Dict[str, List[Tuple[float, Prompt]]] = defaultdict(list)
        fully_evaled_param_combos: Dict[str, Dict[str, Union[Prompt, float]]] = {}

        def objective(trial: optuna.Trial) -> float:
            nonlocal best_prompt, best_score, trial_logs, total_eval_calls, param_score_dict, fully_evaled_param_combos

            logger.info(f"Starting trial num: {trial.number}")
            trial_logs[trial.number] = {}

            candidate_prompt: Prompt = student.deepcopy()

            score_based_instruction_idx: int = trial.suggest_categorical(
                "score_based_instruction", range(len(score_based_instruction_candidates))
            )
            chosen_params: List[int] = [score_based_instruction_idx]

            format_based_instruction_idx: int = trial.suggest_categorical(
                "format_based_instruction", range(len(format_based_instruction_candidates))
            )
            chosen_params.append(format_based_instruction_idx)

            trial_logs[trial.number].update(
                {
                    "score_based_instruction": score_based_instruction_idx,
                    "format_based_instruction": format_based_instruction_idx,
                }
            )

            score_based_candidate_prompt = score_based_instruction_candidates[
                score_based_instruction_idx
            ]
            format_based_candidate_prompt = format_based_instruction_candidates[
                format_based_instruction_idx
            ]

            max_retries = 3
            retry_count = 0

            while retry_count < max_retries:
                try:
                    merged_candidate_prompt_text = run_async(
                        self.merge_prompts(
                            **{
                                "basic_prompt": student.messages,
                                "instruction_improved_prompt": score_based_candidate_prompt,
                                "format_improved_prompt": format_based_candidate_prompt,
                            }
                        )
                    )

                    merged_candidate_prompt = extract_prompt(merged_candidate_prompt_text)
                    merged_candidate_prompt = Prompt.load(merged_candidate_prompt)
                    break
                except Exception as e:
                    retry_count += 1
                    if retry_count == max_retries:
                        raise e
                    logger.warning(f"Attempt {retry_count} failed. Retrying...")

            candidate_prompt.messages = merged_candidate_prompt.messages
            candidate_prompt.fewshot = best_fewshot

            # print(candidate_prompt)

            trial_logs[trial.number]["prompt_path"] = save_candidate_prompt(
                candidate_prompt, log_dir, trial.number
            )

            batch_size: int = self._get_batch_size(minibatch, testset)
            score: float = run_async(
                eval_candidate_prompt(batch_size, testset, candidate_prompt, testset_evaluate)
            )

            categorical_key: str = ",".join(map(str, chosen_params))
            param_score_dict[categorical_key].append((score, candidate_prompt))

            trial_logs[trial.number].update(
                {
                    "num_eval_calls": batch_size,
                    "full_eval": batch_size >= len(testset),
                    "score": score,
                    "pruned": False,
                    "total_eval_calls_so_far": total_eval_calls + batch_size,
                }
            )
            total_eval_calls += batch_size

            if score > best_score:
                print(
                    f"Best Score Updated to Score-based : {score_based_instruction_idx}, Format-based : {format_based_instruction_idx}"
                )
                best_score = score
                best_prompt = candidate_prompt.deepcopy()

            if score >= goal_score:
                trial.study.stop()

            return score

        logger.info("Start Optuna Study to find best combination")

        study: optuna.Study = optuna.create_study(
            direction="maximize",
            sampler=optuna.samplers.TPESampler(seed=seed, multivariate=True),
        )
        study.optimize(objective, n_trials=max_steps)

        if best_prompt is not None and self.track_stats:
            best_prompt.metadata["trial_logs"] = trial_logs
            best_prompt.metadata["score"] = best_score
            best_prompt.metadata["total_eval_calls"] = total_eval_calls

        if log_dir:
            with open(os.path.join(log_dir, "best_prompt.prompt"), "w", encoding="utf-8") as f:
                f.write(best_prompt.dump())

            with open(os.path.join(log_dir, "optuna_study.pkl"), "wb") as file:
                pickle.dump(study, file)

        return best_prompt



import asyncio
import random
from typing import Dict, List, Optional

from ape.common.prompt import Prompt
from ape.common.prompt.utils import format_fewshot
from ape.common.types.response_format import (
    ResponseFormat,
)
from ape.common.utils import logger
from ape.common.types import DatasetItem

from ape.core.core_prompts import ApeCorePrompts
from ape.core.proposer.dataset_summary_generator import create_dataset_summary
from ape.core.proposer.utils import (
    extract_prompt,
    get_response_format_instructions,
)
from ape.core.proposer.propose_base import Proposer

# Hardcoded variables
# MAX_INSTRUCT_IN_HISTORY = 5  # 10

TIPS = {
    "creative": "Don't be afraid to be creative when creating the new instruction!",
    "simple": "Keep the instruction clear and concise.",
    "description": "Make sure your instruction is very informative and descriptive. You can add some hand-crafted examples to help the LLM understand the task better.",
    "high_stakes": "The instruction should include a high stakes scenario in which the LM must solve the task!",
    "persona": 'Include a persona that is relevant to the task in the instruction (ie. "You are a ...")',
}


class GroundedProposer(Proposer):
    """
    A class for generating grounded proposals based on various criteria and configurations.

    This class extends the Proposer base class and provides methods for generating
    instruction proposals using dataset summaries, task demonstrations, instruction history,
    and other configurable parameters.

    Attributes:
        use_dataset_summary (bool): Whether to use dataset summary in proposals.
        use_task_demos (bool): Whether to use task demonstrations in proposals.
        use_tip (bool): Whether to include tips in proposals.
        trainset (List[DatasetItem]): The training dataset.
        view_data_batch_size (int): The batch size for viewing data.
        describe_prompt (Prompt): The prompt for describing tasks.
        generate_instructions (Prompt): The prompt for generating instructions.
        data_summary (Optional[str]): The summary of the dataset.
    """

    def __init__(
        self,
        trainset: List[DatasetItem],
        use_dataset_summary=True,
        use_task_demos=True,
        use_tip=True,
        view_data_batch_size=10,
    ):
        """
        Initialize the GroundedProposer.

        Args:
            trainset (List[DatasetItem]): The training dataset.
            use_dataset_summary (bool, optional): Whether to use dataset summary. Defaults to True.
            use_task_demos (bool, optional): Whether to use task demonstrations. Defaults to True.
            use_tip (bool, optional): Whether to include tips. Defaults to True.
            view_data_batch_size (int, optional): The batch size for viewing data. Defaults to 10.
        """
        self.use_dataset_summary = use_dataset_summary
        self.use_task_demos = use_task_demos
        self.use_tip = use_tip

        self.trainset: List[DatasetItem] = trainset
        self.view_data_batch_size = view_data_batch_size
        self.describe_prompt = ApeCorePrompts.get("describe-prompt")
        self.generate_instructions = ApeCorePrompts.get("gen-instructions")
        self.data_summary = None

    async def prepare_dataset_summary(
        self,
        view_data_batch_size=10,
    ):
        """
        Prepare a summary of the dataset.

        Args:
            view_data_batch_size (int, optional): The batch size for viewing data. Defaults to 10.
        """
        self.view_data_batch_size = view_data_batch_size
        logger.info("Preparing dataset summary")
        self.data_summary = await create_dataset_summary(
            trainset=self.trainset,
            view_data_batch_size=self.view_data_batch_size,
        )
        logger.info(f"DATA SUMMARY: {self.data_summary}")

    async def propose_prompts(
        self,
        N: int,
        T: float,
        task_description: Optional[str] = None,
        prompt_desc: Optional[str] = None,
        base_prompt: Optional[Prompt] = None,
        fewshot_candidates: Optional[List[List[DatasetItem]]] = None,
        inputs_desc: Optional[Dict[str, str]] = None,
        outputs_desc: Optional[Dict[str, str]] = None,
        response_format: Optional[ResponseFormat] = None,
    ) -> List[Prompt]:
        """
        Propose a set of new instructions for the task based on specified criteria.

        Args:
            N (int): Number of proposals to generate.
            T (float): Temperature for generation.
            task_description (Optional[str], optional): Description of the task. Defaults to None.
            prompt_desc (Optional[str], optional): Description of the prompt. Defaults to None.
            base_prompt (Optional[Prompt], optional): Base prompt to start from. Defaults to None.
            fewshot_candidates (Optional[List[List[DatasetItem]]], optional): Candidates for few-shot learning. Defaults to None.
            inputs_desc (Optional[Dict[str, str]], optional): Description of inputs. Defaults to None.
            outputs_desc (Optional[Dict[str, str]], optional): Description of outputs. Defaults to None.
            response_format (Optional[ResponseFormat], optional): Format for the response. Defaults to None.

        Returns:
            List[Prompt]: A list of proposed prompts.
        """
        if not self.data_summary and self.trainset:
            await self.prepare_dataset_summary()

        if not fewshot_candidates:
            fewshot_candidates = [
                random.sample(self.trainset, min(len(self.trainset), 3)) for _ in range(N)
            ]

        proposed_instructions = []

        _tasks = [
            self.propose_one(
                index=i,
                task_description=task_description,
                base_prompt=base_prompt,
                fewshot=fewshot_candidates[i],
                prompt_desc=prompt_desc,
                T=T,
                inputs_desc=inputs_desc,
                outputs_desc=outputs_desc,
                response_format=response_format,
            )
            for i in range(len(fewshot_candidates))
        ]

        proposed_instructions = await asyncio.gather(*_tasks)

        return proposed_instructions

    async def propose_one(
        self,
        T: float,
        index: int,
        task_description: Optional[str] = None,
        prompt_desc: Optional[str] = None,
        base_prompt: Optional[Prompt] = None,
        fewshot: Optional[List[DatasetItem]] = None,
        inputs_desc: Optional[Dict[str, str]] = None,
        outputs_desc: Optional[Dict[str, str]] = None,
        response_format: Optional[ResponseFormat] = None,
        retry_count: int = 0,
    ) -> Prompt:
        """
        Propose a single instruction based on the given criteria.

        Args:
            T (float): Temperature for generation.
            task_description (Optional[str], optional): Description of the task. Defaults to None.
            prompt_desc (Optional[str], optional): Description of the prompt. Defaults to None.
            base_prompt (Optional[Prompt], optional): Base prompt to start from. Defaults to None.
            fewshot (Optional[List[DatasetItem]], optional): Dataset for few-shot learning. Defaults to None.
            inputs_desc (Optional[Dict[str, str]], optional): Description of inputs. Defaults to None.
            outputs_desc (Optional[Dict[str, str]], optional): Description of outputs. Defaults to None.
            response_format (Optional[ResponseFormat], optional): Format for the response. Defaults to None.
            retry_count (int, optional): Number of retry attempts. Defaults to 0.

        Returns:
            Prompt: A proposed prompt.
        """

        if self.use_tip:
            selected_tip = list(TIPS.values())[index % len(TIPS)]
        else:
            selected_tip = ""

        if self.use_task_demos and fewshot:
            logger.info("Formatting fewshot for generation")
            task_fewshot = format_fewshot(fewshot=fewshot, response_format=response_format)
        else:
            task_fewshot = "-"

        if response_format:
            response_format_instructions = get_response_format_instructions(response_format)
        else:
            response_format_instructions = "-"

        output = await self.generate_instructions(
            lm_config=dict(temperature=T),
            task_description=task_description,
            dataset_desc=self.data_summary if self.use_dataset_summary else "-",
            task_fewshot=task_fewshot,
            prompt_desc=prompt_desc if prompt_desc else "-",
            basic_prompt=base_prompt.dump(),
            tip=selected_tip if self.use_tip else "-",
            inputs_desc=inputs_desc if inputs_desc else "-",
            outputs_desc=outputs_desc if outputs_desc else "-",
            response_format_instructions=response_format_instructions,
        )

        try:
            extracted_prompt = extract_prompt(output)
            new_prompt = Prompt.load(extracted_prompt)
            if not new_prompt.messages:
                raise ValueError("Generated prompt has no messages")
            new_prompt.model = base_prompt.model
            new_prompt.name = base_prompt.name
            new_prompt.inputs_desc = inputs_desc
            new_prompt.outputs_desc = outputs_desc
            new_prompt.response_format = response_format

            return new_prompt

        except Exception as e:
            logger.error(f"Error in propose_one: {e}")
            logger.error(f"Output: {output}")
            if retry_count < 3:
                logger.warning(f"Retry attempt {retry_count + 1} for propose_one")
                return await self.propose_one(
                    # trial_logs=trial_logs,
                    index=index,
                    T=T,
                    task_description=task_description,
                    prompt_desc=prompt_desc,
                    base_prompt=base_prompt,
                    fewshot=fewshot,
                    inputs_desc=inputs_desc,
                    outputs_desc=outputs_desc,
                    response_format=response_format,
                    retry_count=retry_count + 1,
                )
            else:
                logger.error("Max retries reached. Returning base prompt.")
                return base_prompt



import json
import random
from typing import Dict, List, Optional, Tuple, Union

import optuna
from pydantic import ConfigDict
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem, ResponseFormat

from ape.core.optimizer.mipro.mipro_base import MIPROBase
from ape.core.optimizer.mipro.mipro_proposer import MIPROProposer
from ape.core.optimizer.utils import reformat_prompt
from ape.core.optimizer import OptunaSingletonStorage


class MIPROWithHIL(MIPROBase):
    """MIPRO optimizer with Human-In-the-Loop capabilities."""

    instruction_candidates: List[Prompt] = []
    fewshot_candidates: List[List[DatasetItem]] = []
    storage: optuna.storages.RDBStorage = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, db_url: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.storage = OptunaSingletonStorage.get_instance(db_url)

    async def create_or_load_study(
        self,
        study_name: str,
        trainset: Optional[List[DatasetItem]] = None,
        task_description: Optional[str] = None,
        prompt_desc: Optional[str] = None,
        inputs_desc: Optional[Dict[str, str]] = None,
        outputs_desc: Optional[Dict[str, str]] = None,
        response_format: Optional[ResponseFormat] = None,
        base_prompt: Optional[Prompt] = None,
        max_demos: int = 5,
    ) -> Tuple[optuna.Study, bool]:
        """
        Create or load an Optuna study for prompt optimization.

        Args:
            study_name (str): Name of the study.
            trainset (Optional[Dataset]): Training dataset.
            task_description (Optional[str]): Description of the task.
            prompt_desc (Optional[str]): Description of the prompt.
            inputs_desc (Optional[Dict[str, str]]): Description of inputs.
            outputs_desc (Optional[Dict[str, str]]): Description of outputs.
            response_format (Optional[ResponseFormat]): Format of the response.
            base_prompt (Optional[Prompt]): Base prompt to start with.
            max_demos (int): Maximum number of demonstrations.

        Returns:
            Tuple[optuna.Study, bool]: The created or loaded study and a boolean indicating if it's a new study.
        """
        if trainset is None:
            trainset = []

        if self.storage is None:
            raise ValueError("Storage is not set up.")

        study = optuna.create_study(
            study_name=study_name,
            storage=self.storage,
            load_if_exists=True,
            direction="maximize",
            sampler=optuna.samplers.TPESampler(multivariate=True),
        )

        is_new_study = len(study.trials) == 0

        if is_new_study:
            # if not trainset:
            #     raise ValueError("Trainset is required for a new study.")
            # if not task_description:
            #     raise ValueError("Task description is required for a new study.")
            if base_prompt and response_format:
                base_prompt = await reformat_prompt(
                    prompt=base_prompt, response_format=response_format
                )

            proposer = MIPROProposer(**self.model_dump())
            self.instruction_candidates = await proposer.generate_candidates(
                prompt=base_prompt,
                trainset=trainset,
                task_description=task_description,
                prompt_desc=prompt_desc,
                inputs_desc=inputs_desc,
                outputs_desc=outputs_desc,
                response_format=response_format,
            )
            study.set_user_attr(
                "prompt_candidates",
                json.dumps([p.dump() for p in self.instruction_candidates]),
            )
            self.fewshot_candidates = [
                [d for d in random.sample(trainset, min(max_demos, len(trainset)))]
                for _ in range(self.num_candidates)
            ]

            study.set_user_attr(
                "fewshot_candidates",
                json.dumps(
                    [
                        [d.model_dump() for d in fewshot_set]
                        for fewshot_set in self.fewshot_candidates
                    ]
                ),
            )
        else:
            prompt_candidates_json = study.user_attrs["prompt_candidates"]
            fewshot_candidates_json = study.user_attrs["fewshot_candidates"]
            self.instruction_candidates = [
                Prompt.load(p) for p in json.loads(prompt_candidates_json)
            ]
            self.fewshot_candidates = json.loads(fewshot_candidates_json)

        return study, is_new_study

    def suggest_next_prompt(self, study: optuna.Study) -> Tuple[optuna.Trial, Prompt]:
        """
        Suggests the next prompt for the optimization study.

        Args:
            study (optuna.Study): The optimization study.

        Returns:
            Tuple[optuna.Trial, Prompt]: A tuple containing the suggested trial and prompt.
        """
        trial = study.ask()
        if len(self.instruction_candidates) == 0:
            raise ValueError("No instruction candidates available.")
        instruction_idx = trial.suggest_categorical(
            "instruction",
            range(len(self.instruction_candidates)),
        )
        prompt = self.instruction_candidates[instruction_idx]
        if len(self.fewshot_candidates) > 0:
            fewshot_idx = trial.suggest_categorical("fewshot", range(len(self.fewshot_candidates)))
            prompt.fewshot = self.fewshot_candidates[fewshot_idx]

        return trial, prompt

    def complete_trial(self, study: optuna.Study, trial: Union[optuna.Trial, int], score: float):
        """
        Completes a trial by telling the study the score obtained.

        Args:
            study (optuna.Study): The study object.
            trial (Union[optuna.Trial, int]): The trial object or trial number.
            score (float): The score obtained for the trial.
        """
        study.tell(trial, score)

    async def get_best_prompt(self, study: optuna.Study) -> Prompt:
        """
        Retrieves the best prompt based on the given Optuna study.

        Args:
            study (optuna.Study): The Optuna study object.

        Returns:
            Prompt: The best prompt object.

        """
        best_trial = study.best_trial
        instruction_idx = best_trial.params["instruction"]
        fewshot_idx = best_trial.params["fewshot"]
        prompt = self.instruction_candidates[instruction_idx]
        prompt.fewshot = self.fewshot_candidates[fewshot_idx]

        return prompt



from .optimizer.mipro.mipro import MIPRO
from .optimizer.mipro.mipro_with_hil import MIPROWithHIL
from .optimizer.mipro.mipro_instruct import MIPROInstruct


__all__ = [
    "MIPRO",
    "MIPROWithHIL",
    "MIPROInstruct",
]



import asyncio
import random
import threading
from typing import Any, Dict, List, Optional, Union
from pydantic import Field, ConfigDict

from ape.common.generate.generate import Generate
from ape.common.generate.generate_base import BaseGenerate
from ape.common.metric.metric_base import BaseMetric
from ape.common.prompt import Prompt
from ape.common.utils import logger
from ape.common.types import DatasetItem

from ape.core.optimizer.sampled_fewshot import SampledFewshot
from ape.core.optimizer.optimizer_base import Optimizer


class BootstrapFewShot(Optimizer):
    """
    A class for optimizing prompts using bootstrapped few-shot examples.

    This optimizer uses a combination of labeled and bootstrapped examples to create
    an optimized prompt. It iteratively improves the prompt by bootstrapping examples
    from the training set.

    Attributes:
        metric (Optional[BaseMetric]): The metric used to evaluate bootstrapped examples.
        metric_threshold (Optional[float]): The threshold for the metric to consider an example successful.
        teacher_settings (Dict): Settings for the teacher prompt.
        max_bootstrapped_demos (int): Maximum number of bootstrapped demonstrations to include.
        max_labeled_demos (int): Maximum number of labeled demonstrations to include.
        max_rounds (int): Maximum number of bootstrapping rounds.
        max_errors (int): Maximum number of errors allowed before raising an exception.
        error_count (int): Current count of errors encountered.
        error_lock (threading.Lock): Lock for thread-safe error counting.
        bootstrapped_fewshot (Optional[Dataset]): Dataset of successfully bootstrapped examples.
        validation (Optional[Dataset]): Dataset used for validation.
    """

    generate: Optional[BaseGenerate] = Generate()
    metric: Optional[BaseMetric] = None
    metric_threshold: Optional[float] = None
    teacher_settings: Dict = {}
    max_bootstrapped_demos: int = 4
    max_labeled_demos: int = 3
    max_rounds: int = 1
    max_errors: int = 5
    error_count: int = 0
    error_lock: threading.Lock = Field(default_factory=threading.Lock, exclude=True)
    bootstrapped_fewshot: Optional[List[DatasetItem]] = None
    validation: Optional[List[DatasetItem]] = None

    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, **data):
        """Initialize the BootstrapFewShot optimizer."""
        super().__init__(**data)
        self.error_lock = threading.Lock()

    async def optimize(
        self,
        student: Prompt,
        *,
        teacher: Optional[Prompt] = None,
        trainset: List[DatasetItem],
        **kwargs,
    ) -> Prompt:
        """
        Optimize the student prompt using bootstrapped few-shot examples.

        Args:
            student (Prompt): The student prompt to be optimized.
            teacher (Optional[Prompt]): The teacher prompt used for bootstrapping.
            trainset (Dataset): The dataset used for training and bootstrapping.

        Returns:
            Prompt: The optimized student prompt.
        """
        self.trainset = trainset

        await self._prepare_student_and_teacher(student, teacher)
        await self._bootstrap()

        self.student = self._train()
        self.student.set_optimized(True)

        return self.student

    async def _prepare_student_and_teacher(self, student: Prompt, teacher: Prompt):
        """
        Prepare the student and teacher prompts for optimization.

        Args:
            student (Prompt): The student prompt to be optimized.
            teacher (Prompt): The teacher prompt used for bootstrapping.
        """
        self.student = student.reset_copy()
        self.teacher = teacher.deepcopy() if teacher is not None else student.reset_copy()

        assert self.student.is_optimized() is False, "Student must be unoptimized."

        if self.max_labeled_demos and self.teacher.is_optimized() is False:
            optimizer = SampledFewshot(k=self.max_labeled_demos)
            self.teacher = await optimizer.optimize(
                self.teacher.reset_copy(), trainset=self.trainset
            )

    async def _bootstrap(self, *, max_bootstraps=None):
        """
        Bootstrap examples from the training set.

        Args:
            max_bootstraps (Optional[int]): Maximum number of examples to bootstrap.
        """
        max_bootstraps = max_bootstraps or self.max_bootstrapped_demos

        bootstrapped = {}
        self.bootstrapped_fewshot = []

        for round_idx in range(self.max_rounds):
            tasks = []
            for example_idx, example in enumerate(self.trainset):
                if len(bootstrapped) >= max_bootstraps:
                    break

                if example_idx not in bootstrapped:
                    tasks.append(self._bootstrap_one_example(example, round_idx))

                if len(tasks) == 25 or example_idx == len(self.trainset) - 1:
                    results = await asyncio.gather(*tasks)
                    for idx, success in enumerate(results):
                        if success:
                            bootstrapped[example_idx - len(tasks) + 1 + idx] = True
                    tasks = []

            if len(bootstrapped) >= max_bootstraps:
                break

        logger.info(
            f"Bootstrapped {len(bootstrapped)} full traces after {len(self.trainset)} examples in round {round_idx}.",
        )

        self.validation = [x for idx, x in enumerate(self.trainset) if idx not in bootstrapped]
        random.Random(0).shuffle(self.validation)

        # NOTE: Can't yet use evaluate because we need to trace *per example*
        # evaluate = Evaluate(program=self.teacher, metric=self.metric, num_threads=12)
        # score = evaluate(self.metric, display_table=False, display_progress=True)

    async def _bootstrap_one_example(self, example: DatasetItem, round_idx=0):
        """
        Bootstrap a single example from the training set.

        Args:
            example (DatasetItem): The example to bootstrap.
            round_idx (int): The current bootstrapping round index.

        Returns:
            bool: True if the example was successfully bootstrapped, False otherwise.
        """
        teacher = self.teacher
        cache = []

        try:
            lm_config = (
                {"temperature": 0.7 + 0.001 * round_idx} if round_idx > 0 else {"temperature": 1.0}
            )

            cache = teacher.fewshot
            teacher.fewshot = [x for x in teacher.fewshot if x != example]
            inputs = example.get("inputs", {})
            outputs = example.get("outputs", {})
            prediction: Union[str, Dict[str, Any]] = await self.generate(
                teacher,
                inputs,
                lm_config=lm_config,
            )
            teacher.fewshot = cache

            if self.metric is not None:
                metric_val = await self.metric(inputs=inputs, gold=outputs, pred=prediction)
                if self.metric_threshold:
                    success = metric_val.score >= self.metric_threshold
                else:
                    success = metric_val.score
            else:
                success = True
        except Exception as e:
            success = False
            with self.error_lock:
                self.error_count += 1
                current_error_count = self.error_count
            if current_error_count >= self.max_errors:
                raise e
            logger.error(f"Failed to run or to evaluate example.\n Error: {e}.")

        if success:
            self.bootstrapped_fewshot.append(example)

        return success

    def _train(self):
        """
        Train the student prompt using the bootstrapped and raw few-shot examples.

        Returns:
            Prompt: The trained student prompt.
        """
        rng = random.Random(0)
        raw_fewshot = self.validation
        augmented_fewshot = self.bootstrapped_fewshot[: self.max_bootstrapped_demos]

        sample_size = min(self.max_labeled_demos - len(augmented_fewshot), len(raw_fewshot))
        sample_size = max(sample_size, 0)

        raw_fewshot = rng.sample(raw_fewshot, sample_size)

        self.student.fewshot = raw_fewshot + augmented_fewshot

        return self.student



from typing import List, Optional
from pydantic import BaseModel
from abc import ABC, abstractmethod
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem


class Optimizer(BaseModel, ABC):
    """
    Abstract base class for prompt optimization algorithms.

    Attributes:
        student (Optional[Prompt]): The student prompt to be optimized.
        teacher (Optional[Prompt]): An optional teacher prompt used in some optimization strategies.
        trainset (Optional[List[DatasetItem]]): The dataset used for training/optimization.
    """

    student: Optional[Prompt] = None
    teacher: Optional[Prompt] = None
    trainset: Optional[List[DatasetItem]] = None

    @abstractmethod
    async def optimize(self, student: Prompt, *, trainset: List[DatasetItem], **kwargs) -> Prompt:
        """
        Abstract method to optimize a given prompt.

        Args:
            student (Prompt): The initial prompt to be optimized.
            trainset (List[DatasetItem]): The dataset used for optimization.
            **kwargs: Additional keyword arguments specific to the optimization algorithm.

        Returns:
            Prompt: The optimized prompt.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")



import threading
import sqlalchemy
import optuna


class OptunaSingletonStorage:
    """
    A singleton class for managing Optuna storage.

    This class ensures that only one instance of the Optuna storage is created
    and reused across the application, following the singleton pattern.
    """

    _instance = None
    _lock = threading.Lock()

    def __init__(self):
        """
        Private constructor to prevent direct instantiation.
        Raises a RuntimeError if called directly.
        """
        raise RuntimeError("Call get_instance() instead")

    @classmethod
    def get_instance(cls, db_url):
        """
        Get or create the singleton instance of OptunaSingletonStorage.

        Args:
            db_url (str): The database URL for creating the storage.

        Returns:
            optuna.storages.RDBStorage: The singleton instance of Optuna RDB storage.
        """
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = cls._create_storage(db_url)
        return cls._instance

    @staticmethod
    def _create_storage(db_url):
        """
        Create and configure the Optuna RDB storage.

        This method creates a SQLAlchemy engine, ensures the 'optuna' schema exists,
        and sets up the Optuna RDB storage with specific configuration.

        Args:
            db_url (str): The database URL for creating the storage.

        Returns:
            optuna.storages.RDBStorage: Configured Optuna RDB storage instance.
        """
        engine = sqlalchemy.create_engine(db_url)

        with engine.connect() as connection:
            connection.execute(sqlalchemy.text("CREATE SCHEMA IF NOT EXISTS optuna;"))
            connection.commit()

        # Create Optuna RDB storage with schema specified
        storage_url = f"{db_url}?options=-c%20search_path=optuna"
        return optuna.storages.RDBStorage(
            url=storage_url,
            engine_kwargs={
                "pool_size": 20,
                "max_overflow": 0,
                "connect_args": {"options": "-c search_path=optuna"},
            },
        )



from typing import Dict, List, Optional
from ape.common.prompt import Prompt
from ape.common.types import DatasetItem, ResponseFormat
from ape.common.utils import logger

from ape.core.optimizer.mipro.mipro_base import MIPROBase
from ape.core.proposer.grounded_proposer import GroundedProposer


class MIPROProposer(MIPROBase):
    """
    A class for generating prompt candidates using the MIPRO (Model-based Instruction Prompt Optimization) approach.
    """

    async def generate_candidates(
        self,
        trainset: List[DatasetItem],
        task_description: Optional[str] = None,
        prompt_desc: Optional[str] = None,
        prompt: Optional[Prompt] = None,
        fewshot_candidates: Optional[List[List[DatasetItem]]] = None,
        inputs_desc: Optional[Dict[str, str]] = None,
        outputs_desc: Optional[Dict[str, str]] = None,
        response_format: Optional[ResponseFormat] = None,
    ) -> List[Prompt]:
        """
        Generate a list of prompt candidates based on the given parameters.

        Args:
            trainset (List[DatasetItem]): The dataset used for training.
            task_description (Optional[str]): A description of the task.
            prompt_desc (Optional[str]): A description of the prompt.
            prompt (Optional[Prompt]): A base prompt to start from.
            fewshot_candidates (Optional[List[List[DatasetItem]]]): A list of few-shot example datasets.
            inputs_desc (Optional[Dict[str, str]]): A description of the input fields.
            outputs_desc (Optional[Dict[str, str]]): A description of the output fields.
            response_format (Optional[ResponseFormat]): The desired format for the response.

        Returns:
            List[Prompt]: A list of generated prompt candidates.
        """
        logger.info("Initializing GroundedProposer")
        proposer = GroundedProposer(
            trainset=trainset,
            view_data_batch_size=self.view_data_batch_size,
        )

        proposer.use_tip = True

        logger.info(f"Generating {self.num_candidates} instruction candidates")
        prompt_candidates = await proposer.propose_prompts(
            prompt_desc=prompt_desc,
            base_prompt=prompt,
            fewshot_candidates=fewshot_candidates,
            N=self.num_candidates,
            T=self.init_temperature,
            task_description=task_description,
            inputs_desc=inputs_desc,
            outputs_desc=outputs_desc,
            response_format=response_format,
        )

        return prompt_candidates



import json
import re
from typing import Any, Dict, List

from pydantic import BaseModel
from ape.common.prompt import Prompt
from ape.common.types import ResponseFormat


def extract_prompt(text: str) -> str:
    match = re.search(r"```prompt(.*?)```", text, re.DOTALL)
    if match:
        return match.group(1).strip()
    else:
        raise ValueError("No prompt found")


def get_response_format_instructions(response_format: ResponseFormat):
    if response_format["type"] == "json_object":
        return (
            "The prompt should enforce a JSON output and must include the word JSON in the prompt."
        )
    else:
        return ""


class HistoryItem(BaseModel):
    prompt: Prompt
    score: float


class AggregatedHistoryItem(BaseModel):
    total_score: float
    count: int


def create_history_string(prompt: Prompt, trial_logs: Dict[str, Any], top_n: int):
    # TODO: NOT: Fix
    instruction_aggregate: Dict[str, AggregatedHistoryItem] = {}
    instruction_history: List[HistoryItem] = []

    # Load trial programs
    for trial_num in trial_logs:
        trial = trial_logs[trial_num]
        if "prompt_path" in trial:
            trial_prompt = prompt.deepcopy()
            # TODO: NOTE: look into this
            trial_prompt.load(trial["prompt_path"])
            instruction_history.append(
                HistoryItem(
                    prompt=trial_prompt,
                    score=trial["score"],
                )
            )

    # Aggregate scores for each instruction
    for history_item in instruction_history:
        instruction: str = json.dumps(history_item.prompt.messages)
        score = history_item.score

        if instruction in instruction_aggregate:
            instruction_aggregate[instruction].total_score += score
            instruction_aggregate[instruction].count += 1
        else:
            instruction_aggregate[instruction] = AggregatedHistoryItem(
                total_score=score,
                count=1,
            )

    # Calculate average score for each instruction and prepare for sorting
    predictor_history = []
    for instruction, data in instruction_aggregate.items():
        average_score = data.total_score / data.count
        predictor_history.append((instruction, average_score))

    # Deduplicate and sort by average score, then select top N
    seen_instructions = set()
    unique_predictor_history = []
    for instruction, score in predictor_history:
        if instruction not in seen_instructions:
            seen_instructions.add(instruction)
            unique_predictor_history.append((instruction, score))

    top_instructions = sorted(unique_predictor_history, key=lambda x: x[1], reverse=True)[:top_n]
    top_instructions.reverse()

    # Create formatted history string
    predictor_history_string = ""
    for instruction, score in top_instructions:
        predictor_history_string += instruction + f" | Score: {score}\n\n"

    return predictor_history_string



from pydantic import BaseModel
from typing import Dict, Optional

from ape.common.evaluate.evaluate import Evaluate
from ape.common.generate.generate import Generate
from ape.common.generate.generate_base import BaseGenerate
from ape.common.metric import BaseMetric
from ape.common.global_metric import BaseGlobalMetric


class MIPROBase(BaseModel):
    """
    Base class for MIPRO (Model-based Instruction Prompt Optimization) implementations.

    Attributes:
        task_model (Optional[str]): The model used for executing the task. Default is "gpt-4o-mini".
        teacher_settings (Dict): Settings for the teacher model. Default is an empty dictionary.
        num_candidates (int): Number of candidate prompts to generate. Default is 10.
        metric (Optional[BaseMetric]): The metric object used for evaluation. Default is None.
        init_temperature (float): Initial temperature for sampling. Default is 1.0.
        verbose (bool): Whether to print verbose output. Default is False.
        track_stats (bool): Whether to track statistics during optimization. Default is True.
        log_dir (Optional[str]): Directory for logging. Default is None.
        view_data_batch_size (int): Batch size for viewing data. Default is 10.
        minibatch_size (int): Size of minibatches for evaluation. Default is 25.
        update_prompt_after_full_eval (bool): Whether to only update the best prompt after full evaluation. Default is True.
        minibatch_full_eval_steps (int): Number of steps for full evaluation on minibatches. Default is 10.
    """

    task_model: Optional[str] = "gpt-4o-mini"
    teacher_settings: Dict = {}
    num_candidates: int = 10
    generate: Optional[BaseGenerate] = Generate()
    evaluate: Optional[Evaluate] = None  # If provided, 'metric' and 'global_metric' will be ignored
    metric: Optional[BaseMetric] = None
    global_metric: Optional[BaseGlobalMetric] = None
    init_temperature: float = 1.0
    verbose: bool = False
    track_stats: bool = True
    log_dir: Optional[str] = None
    view_data_batch_size: int = 10
    minibatch_size: int = 25
    update_prompt_after_full_eval: bool = True
    minibatch_full_eval_steps: int = 10

    class Config:
        arbitrary_types_allowed = True
